// This is a sketch for an Arduino Mega to program, verify, and dump data 
// with a 4 megabit SST 39SF040 flash memory chip
// runs at 115200 baud and uses an Intel Hex file to program Flash
// sent to MEGA with ascii file transfer
//    to begin ascii file transfer in minicom: cntrl a, s, select "ascii" from
//    pulldown menu use down arrow to find file, select with spacebar, press enter

#define ce 39   // PORTG 2
#define oe 40   // PORTG 1
#define we 41   // PORTG 0

byte lowaddress = 0x00;   // PORTC
byte highaddress = 0x00;  // PORTL
byte extaddress = 0x00;   // PORTB
char ch = 0;
char veriflag = 0;
byte hex;
char count;    // has to be signed
byte pos = 0;
char menu[] = {"\n\r Press 'p' to Program, 'v' to Verify, 'r' to Read, 'E' to erase the 4 meg Flash"};
char begintransfer[] = {"\n\rBegin .hex file Ascii transfer - press Esc when done "};


void setup()
{
  Serial.begin(115200);
 
  DDRA = 0x00;   // set PORTA data pins to input
  DDRC = 0xff;   // set PORTC low address pins to output
  DDRL = 0xff;   // set PORTL high address pins to output
  DDRB = 0x07;   // set PORTB ext address pins 0-2 to output
  DDRG = 0x07;   // set PORTG pins 0-2 to output
  
  digitalWrite( ce, HIGH ); // PORTG-2  -CE 
  digitalWrite( oe, HIGH ); // PORTG-1  -OE
  digitalWrite( we, HIGH ); // PORTG-0  -WE  

  Serial.println(menu);
}


void loop()
{
  while(!Serial.available());   // wait for a keypress
 
  ch = Serial.read();           // read key pressed

  if(ch == 'p')
    program();
  if(ch == 'v')
    verify();
  if(ch == 'r')
    readprom();
  if(ch == 'E')
    chiperase();  
}


void progpulse()   // used by unprotect() and chiperase()
{
  digitalWrite( we, LOW );     // write enable low
  digitalWrite( ce, LOW );      // chip select low
                               // 40 ns low needed   
  digitalWrite( ce, HIGH );    // chip select HIGH
  digitalWrite( we, HIGH );    // write enable HIGH
                               
}


void chiperase()
{
  digitalWrite(ce, HIGH);
  digitalWrite(oe, HIGH);
  digitalWrite(we, HIGH);
  
  for(byte a = 22; a < 30; a ++)  
    pinMode( a, OUTPUT);           // change PORTA data pins to output
 
  Serial.println("Erasing EEPROM");
    
  PORTA = 0xaa; // data            // 1. write 0xaa 
  PORTL = 0x55; // high address    // to address 0x5555
  PORTC = 0x55; // low address   
  progpulse();

  PORTA = 0x55; // data            // 2. write 0x55 
  PORTL = 0x2a; // high address    // to address 0x2aaa
  PORTC = 0xaa; // low address 
  progpulse();
  
  PORTA = 0x80; // data            // 3. write 0x80 
  PORTL = 0x55; // high address    // to address 0x5555
  PORTC = 0x55; // low address 
  progpulse();
  
  PORTA = 0xaa; // data            // 4. write 0xaa 
  PORTL = 0x55; // high address    // to address 0x5555
  PORTC = 0x55; // low address 
  progpulse();
 
  PORTA = 0x55; // data            // 5. write 0x55 
  PORTL = 0x2a; // high address    // to address 0x2aaa
  PORTC = 0xaa; // low address
  progpulse();
 
  PORTA = 0x10; // data            // 6. write 0x20 
  PORTL = 0x55; // high address    // to address 0x5555
  PORTC = 0x55; // low address 
  progpulse();
  
  for(byte a = 22; a < 30; a ++)  
    pinMode( a, INPUT);      // change PORTA data pins to input
 
  Serial.println("\n\rEEPROM erase complete");
  Serial.println(menu);  
}

 
void program()
{
  ch = 0;
  digitalWrite( ce, HIGH );     // chip select HIGH
  digitalWrite( oe, HIGH );     // output enable HIGH
  digitalWrite( we, HIGH );     // write enable HIGH
 
  for(byte a = 22; a < 30; a ++)  
    pinMode( a, OUTPUT);      // change data pins to output
 
  lowaddress = 0x00;
  highaddress = 0x00; 
  extaddress = 0x00;
  PORTL = highaddress;     // high address  
  PORTC = lowaddress;      // low address  
  PORTB = extaddress;      // extended address
  
  pos = 0;
   
  Serial.println(begintransfer);
  while(ch != 0x1b)
    {
    if(Serial.available() > 0)  
      {
      ch = Serial.read(); 
      translate(ch);  
      }
    }  
    
  for(byte a = 22; a < 30; a ++)  
    pinMode( a, INPUT );       // change data pins to input  
 
  digitalWrite( ce, HIGH );
  digitalWrite( oe, HIGH );
  digitalWrite( we, HIGH );
 
  Serial.println("\n\rprogramming complete ");
  Serial.println(menu); 
}


void translate(char dat)     // convert ascii INTEL .hex file data to binary
{ 
  pos ++;
  if(dat == ':')             // start of header bytes
    pos = 0;
  
  if(dat > 0x40)        // A is 0x41, a is 0x61  
    dat = dat + 9;      // add 9 to convert to hexadecimal                       
  dat = (0x0f & dat);   // strip the left digit

  if(pos == 1) 
    count = (dat << 4); // count is converted to use
  if(pos == 2)          // to identify data positions
    count = (dat | count);
    
  // -------------- address translate
 
  
  if(pos == 3 )   // check even/odd position
     hex = (dat << 4);     // odd = left nibble bits 4-7
  if(pos == 4)                    // low nibble bits 0-3  
     highaddress = (hex | dat);  
  if(pos == 5 )   // check even/odd position
     hex = (dat << 4);     // odd = left nibble bits 4-7
  if(pos == 6)                    // low nibble bits 0-3  
     lowaddress = (hex | dat);      
    
  // ------------- end address translate
 
  //  pos 7, 8, and checksum are skipped over
 
  if(pos > 8)
    {
    if((pos & 0x01) == 1)   // check even/odd position
      hex = (dat << 4);     // odd = left nibble bits 4-7
    else                    // low nibble bits 0-3  
      if(count-- > 0)          
        programbyte((hex | dat));
     }  
}


void programbyte( byte dataout )
{                     
  PORTG |= 0x07; // ce, oe, we
  PORTB = 0x00;  // ext address
                 // enter 3 byte unprotect sequence
  PORTA = 0xaa;  // data            // 1. write 0xaa 
  PORTL = 0x55;  // high address    // to address 0x5555
  PORTC = 0x55;  // low address   
  
  
  PORTG &= 0xfa;  // xxxx x010 ce oe we: c=low we=low
  PORTG |= 0x07;  // xxxx x111 ce oe we
 
  PORTA = 0x55; // data   
  PORTL = 0x2a; // high address    // to address 0x5555
  PORTC = 0xaa; // low address 
           
  PORTG &= 0xfa;  // xxxx x010 ce oe we: c=low we=low
  PORTG |= 0x07;  // xxxx x111 ce oe we
  
  PORTA = 0xa0; // data            // 2. write 0x55 
  PORTL = 0x55; // high address    // to address 0x2aaa
  PORTC = 0x55; // low address 
 
  PORTG &= 0xfa;  // xxxx x010 ce oe we: c=low we=low
  PORTG |= 0x07;  // xxxx x111 ce oe we
  
  PORTA = dataout;         // data  
  PORTL = highaddress;     // high address  
  PORTC = lowaddress;      // low address  
  PORTB = extaddress;
  
  PORTG &= 0xfa;  // xxxx x010 ce oe we: c=low we=low
  PORTG |= 0x07;  // xxxx x111 ce oe we
  
  lowaddress++;
  if( lowaddress == 0)
    {
    highaddress++;  
    if(highaddress == 0x00)
       {
       extaddress++;
       if(extaddress == 0x08)  
          ch = 0x1b;          // ch = 0x1b to stop programming
       }
    }   
}


void verify()
{    
  digitalWrite(ce, HIGH);  // chip select HIGH
  digitalWrite(oe, HIGH);  // output enable HIGH
  digitalWrite(we, HIGH);  // write enable HIGH
 
  for(byte a = 22; a < 30; a ++)  
    pinMode( a, INPUT);         // change data pins to input
  
  lowaddress = 0x00;
  highaddress = 0x00; 
  extaddress = 0;
  PORTL = highaddress;     // high address  
  PORTC = lowaddress;      // low address  
  PORTB = extaddress;
  pos = 0;
  veriflag = 0; 
  
  Serial.println(begintransfer);
  while(ch != 0x1b)             // if Esc key pressed after transfer is done, then exit
    {                           // keypress's are ignored while file is transfering
    if(Serial.available() > 0)  // if a file character is sent
      {                         // or key is pressed after file transfer has ended
      ch = Serial.read();       // read it into ch
      verifytranslate(ch);  
      }
    }                           
  if(veriflag == 1)                     
    Serial.print("Verify Failed");
  else
    Serial.println("\n\rVerify Passed");
    
  Serial.println(menu);  
}


void verifytranslate(char dat)
{ 
  pos ++;
  if(dat == ':')        // start of header bytes
    pos = 0;
 
  if(dat > 0x40)        // A is 0x41, a is 0x61
      dat = dat + 9;    // add 9 to convert to hexadecimal                       
  dat = (0x0f & dat);   // strip the left digit
 
  if(pos == 1) 
    count = (dat << 4); // count is converted to use
  if(pos == 2)          // to identify data positions
    count = dat | count;
 
  // address- pos 3, 4, 5, 6, type- pos 7, 8, and checksum are skipped over
  // -------------- address translate
  
  if(pos == 3 )   // check even/odd position
     hex = (dat << 4);     // odd = left nibble bits 4-7
  if(pos == 4)                    // low nibble bits 0-3  
     highaddress = (hex | dat);  
  if(pos == 5 )   // check even/odd position
     hex = (dat << 4);     // odd = left nibble bits 4-7
  if(pos == 6)                    // low nibble bits 0-3  
     lowaddress = (hex | dat);      
 
  // ------------- end address translate
 
  if(pos > 8)
    {
    if((pos & 0x01) == 1)  // check even/odd position
      hex = (dat << 4);    // odd = left nibble bits 4-7
    else                   // even = right nibble bits 0-3  
      if(count-- > 0)          
        verifybyte((hex | dat));
    }   
}


void verifybyte(byte datain)
{  
  byte dataver;
   
  PORTC = lowaddress;     // high address
  PORTL = highaddress;    // low address 
   
  digitalWrite(we, HIGH); // write enable HIGH
  digitalWrite(ce,LOW);   // chip select low
  digitalWrite(oe,LOW);   // chip select low
  delayMicroseconds(1);
  dataver = PINA;         // PINA reads port A as input
  digitalWrite(ce,HIGH);  // chip select HIGH
  digitalWrite(oe, HIGH); // write enable HIGH
 
  lowaddress++;
  if( lowaddress == 0)
    {
    highaddress++;  
    if(highaddress == 0x00)
       {
       extaddress++;
       if(extaddress == 0x08)  
          ch = 0x1b;          // ch = 0x1b to stop programming whi
       }
    }   
 
  if(datain != dataver)  
    veriflag = 1;         // veriflag = 1 is verify failed
}


void readprom()
{
  byte dat;
  byte row = 0;
   
  digitalWrite(ce, HIGH);  // chip select HIGH
  digitalWrite(oe, HIGH);  // output enable HIGH
  digitalWrite(we, HIGH);  // write enable HIGH
 
  ch = 0;
   
  lowaddress = 0;
  highaddress = 0;
  extaddress = 0;
  PORTL = highaddress;     // high address  
  PORTC = lowaddress;      // low address  
  PORTB = extaddress;
  
  DDRA = 0x00;    // change data pins to input

  while(extaddress < 0x08) // 0x80 is end of 32k EEPROM
    {
    PORTC = lowaddress;     // high address
    PORTL = highaddress;    // low address 
    
    // print starting address of 16 byte line
    
    if( row == 0)
      {
      Serial.print("\n\r");
      if(highaddress < 0x10)  
         Serial.print("0");
      Serial.print(highaddress, HEX);
      if(lowaddress < 0x10)  
         Serial.print("0");
      Serial.print(lowaddress, HEX);
      }
    // -- read and display byte read
    
    digitalWrite(we, HIGH); // write enable HIGH
    digitalWrite(ce,LOW);   // chip select LOW
    digitalWrite(oe,LOW);   // output enable LOW
  //  delayMicroseconds(3);
    dat = PINA;
    digitalWrite(ce,HIGH);  // chip select HIGH
    digitalWrite(oe, HIGH); // output enable HIGH
    Serial.print(" ");
    if(dat < 0x10)
       Serial.print("0");
    Serial.print(dat, HEX);
    
    lowaddress++;
    if( lowaddress == 0 )
      {
      highaddress ++; 
      if(highaddress == 0)
         extaddress++;
      }   
    
    row ++;     
    if(row == 16)
       row = 0;
       
    if(Serial.available() > 0)  // if a key is pressed during dump read it
      if(Serial.read() == 0x1b) // if it is Esc - stop the dump
        break; 
    }
  Serial.println("\n\rEEPROM read complete");
  Serial.println(menu);  
}
